LANG_ALIASES = {
    "english": "eng_Latn",
    "eng": "eng_Latn",
    "en": "eng_Latn",
    "ko": "kor_Hang",
    "korean": "kor_Hang",
    "kor": "kor_Hang",
    "kr": "kor_Hang"
}
LEVEL = {
    "fine": "fine",
    "fine-tuned": "fine",
    "origin": "origin",
    "0": "origin",
    "1": "fine"
}
STYLE = {
    "formal":"문어체",
    "informal":"구어체",
    "android":"안드로이드",
    "azae":"아재",
    "chat":"채팅",
    "choding":"초등학생",
    "emoticon":"이모티콘",
    "enfp":"enfp",
    "gentle":"신사",
    "halbae":"할아버지",
    "halmae":"할머니",
    "joongding":"중학생",
    "king":"왕",
    "naruto":"나루토",
    "seonbi":"선비",
    "sosim":"소심한",
    "translator":"번역기",
}


from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
from konlpy.tag import Twitter
import numpy as np

class KoTANAugmentationFactory:
    """
    Text augmentation using facebook/nllb-200-distilled-600M Meta model

    - dataset: Train

    Args:
        src (str): source language
        
    Returns:
        class: KoTANAugmentation class

    Examples:
        >>> mt = KoTAN(task="augmentation", level="fine")
    """

    def __init__(self, 
                 task, 
                 tgt, 
                 level,
                 style
                 ):
        super().__init__()
        self.task = task
        self.tgt = tgt
        self.level = level

    def load(self, device):
        if self.level == "fine":
            ko2en_tokenizer = AutoTokenizer.from_pretrained("NHNDQ/nllb-finetuned-ko2en")
            en2ko_tokenizer = AutoTokenizer.from_pretrained("NHNDQ/nllb-finetuned-en2ko")

            ko2en_model = AutoModelForSeq2SeqLM.from_pretrained("NHNDQ/nllb-finetuned-ko2en").to(device)
            en2ko_model = AutoModelForSeq2SeqLM.from_pretrained("NHNDQ/nllb-finetuned-en2ko").to(device)

        
        elif self.level == "origin":
            ko2en_tokenizer = AutoTokenizer.from_pretrained("facebook/nllb-200-distilled-600M", src_lang="kor_Hang", tgt_lang="eng_Latn")
            en2ko_tokenizer = AutoTokenizer.from_pretrained("facebook/nllb-200-distilled-600M", src_lang="eng_Latn", tgt_lang="kor_Hang")

            ko2en_model = AutoModelForSeq2SeqLM.from_pretrained("facebook/nllb-200-distilled-600M").to(device)
            en2ko_model = AutoModelForSeq2SeqLM.from_pretrained("facebook/nllb-200-distilled-600M").to(device)
        

        return KoTANAugmentation(
            ko2en_tokenizer, en2ko_tokenizer,
            ko2en_model, en2ko_model, 
            device
        )


class KoTANAugmentation:
    def __init__(self, 
                 ko2en_tokenizer, en2ko_tokenizer,
                 ko2en_model, en2ko_model, 
                 device):
        self.ko2en_tokenizer = ko2en_tokenizer
        self.en2ko_tokenizer = en2ko_tokenizer

        self.ko2en_model = ko2en_model
        self.en2ko_model = en2ko_model

        self.device = device

    def predict(self, text):
        """
        Predict a backtranslation result

        Args:
            text (str): input text

        Returns:
            output (list): Backtranslation results
        """       

        if type(text) == type('text'):
            text = [text]

        # post-processing
        post_process_output = []
        
        for i in range(len(text)):
            emojiList, textList = self._post_process(text[i])
            # ko2en
            translation = self._translate(textList, 'eng_Latn', self.ko2en_tokenizer, self.ko2en_model)
            # en2ko
            augList = self._translate(translation, 'kor_Hang', self.en2ko_tokenizer, self.en2ko_model)
            outList = []
            for emo, txt in zip(emojiList, augList):
                output = txt + emo
                outList.append(output)
                output = ''.join(outList).strip()
            post_process_output.append(output)

        return post_process_output


    def _translate(self, text, tgt, tokenizer, model):

        # translation
        inputs = tokenizer(text, padding=True, truncation=True, return_tensors="pt")
        inputs = inputs.to(self.device)
        translated_tokens = model.generate(
            inputs['input_ids'], forced_bos_token_id=tokenizer.lang_code_to_id[tgt], max_length=128
        )

        translation = tokenizer.batch_decode(translated_tokens, skip_special_tokens=True)

        post_output = []

        for inp, outp in zip(text, translation):
            if inp == '':
                post_output.append('')
            else:
                post_output.append(outp)
        
        return post_output

    
    def _post_process(self, text):
        textList = []
        emojiList = []
        twit = Twitter()

        posText = twit.pos(text)
        posArray = np.array(posText)

        for i in range(len(posArray)):
            if posArray[i][1] == 'KoreanParticle':
                emojiList.append(posArray[i][0])

        for i in range(len(emojiList)):
            splitText = text.split(emojiList[i], maxsplit=1)

            if splitText[0] == '':
                textList.append('')
            else:
                textList.append(splitText[0])

            try:
                if len(splitText[1:]) > 1:
                    text = ''.join(splitText[1:]).strip()
                else:
                    text = splitText[1:][0].strip()

            except:
                break

            try:
                if text in emojiList[i+1]:
                    pass
            except:
                textList.append(splitText[-1])
                emojiList.append('')
                break

        ## 이모지 없는 경우            
        if len(emojiList) < 1:
            emojiList.append('')
            textList.append(text)
                
        return emojiList, textList


import torch

from .tasks import (
    KoTANTranslationFactory,
    KoTANAugmentationFactory,
    KoTANStyleConversiontFactory
)

from .const import LANG_ALIASES, LEVEL, STYLE


# Task list
SUPPORTED_TASKS = {
    "mt": KoTANTranslationFactory,
    "translation": KoTANTranslationFactory,
    "aug": KoTANAugmentationFactory,
    "augmentation": KoTANAugmentationFactory,
    "style": KoTANStyleConversiontFactory,
    "convert": KoTANStyleConversiontFactory
}


class KoTAN:
    def __new__(
            cls,
            task,
            tgt="en",
            level="fine",
            style="formal"
            ):
        
        if task not in SUPPORTED_TASKS:
            raise KeyError("Unknown task {}, available tasks are {}".format(
                task,
                list(SUPPORTED_TASKS.keys()),
            ))

        if tgt not in LANG_ALIASES:
            raise KeyError("Unknown target language {}, available target languages are {}".format(
                task,
                list(LANG_ALIASES.keys()),
            ))
        
        if level not in LEVEL:
            raise KeyError("Unknown level {}, available levels are {}".format(
                task,
                list(LEVEL.keys()),
            ))
        
        if style not in STYLE:
            raise KeyError("Unknown style {}, available levels are {}".format(
                task,
                list(STYLE.keys()),
            ))
        
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        task_module = SUPPORTED_TASKS[task](
            task,
            LANG_ALIASES[tgt],
            LEVEL[level],
            STYLE[style]
        ).load(device)

        return task_module

    @staticmethod
    def available_tasks():
        """
        Returns available tasks in KoTAN project

        Returns:
            str: Supported task names

        """
        return "Available tasks are {}".format(list(SUPPORTED_TASKS.keys()))
    
    @staticmethod
    def available_lang():
        """
        Returns available language in KoTAN project

        Returns:
            str: Supported language names

        """
        return "Available tasks are {}".format(list(LANG_ALIASES.keys()))
    
    @staticmethod
    def available_level():
        """
        Returns available level in KoTAN project

        Returns:
            str: Supported level names

        """
        return "Available tasks are {}".format(list(LEVEL.keys()))
    
    @staticmethod
    def available_style():
        """
        Returns available convert style in KoTAN project

        Returns:
            str: Supported style names

        """
        return "Available tasks are {}".format(list(STYLE.keys()))


from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
from konlpy.tag import Twitter
import numpy as np

class KoTANTranslationFactory:
    """
    Machine translation using facebook/nllb-200-distilled-600M Meta model

    - dataset: Train

    Args:
        src (str): source language
        
    Returns:
        class: KoTANTranslation class

    Examples:
        >>> mt = KoTAN(task="translation", tgt="en")
    """
    def __init__(
            self,
            task,
            tgt,
            level,
            style
            ):
        super().__init__()
        self.task = task
        self.tgt = tgt

    def load(self, device: str):
        if self.tgt == "kor_Hang":
            tokenizer = AutoTokenizer.from_pretrained("NHNDQ/nllb-finetuned-en2ko")
            model = AutoModelForSeq2SeqLM.from_pretrained("NHNDQ/nllb-finetuned-en2ko").to(device)
            
        if self.tgt == "eng_Latn":
            tokenizer = AutoTokenizer.from_pretrained("NHNDQ/nllb-finetuned-ko2en")
            model = AutoModelForSeq2SeqLM.from_pretrained("NHNDQ/nllb-finetuned-ko2en").to(device)

        return KoTANTranslation(
            model,
            tokenizer,
            device,
            self.tgt
        )
    

class KoTANTranslation:
    def __init__(self, model, tokenizer, device, tgt):
        self.model = model
        self.tokenizer = tokenizer
        self.device = device
        self.tgt = tgt

    def predict(self, text):
        """
        Predict a translation result

        Args:
            text (str): input text

        Returns:
            output (list): Translation results
        """

        ## text가 '' 일 경우, '' 그대로 출력되도록 후처리
        inputs = self.tokenizer(text, padding=True, truncation=True, return_tensors="pt")
            
        translated_tokens = self.model.generate(
            **inputs.to(self.device), forced_bos_token_id=self.tokenizer.lang_code_to_id[self.tgt], max_length=128
        )

        output = self.tokenizer.batch_decode(translated_tokens, skip_special_tokens=True)

        post_output = []

        for inp, outp in zip(text, output):
            if (inp == '') or (inp == '!') or (inp == '?') or (inp == '.') or (inp == ',') or (inp == ' '):
                post_output.append(inp)
            else:
                post_output.append(outp)

        return post_output
    
    def _post_process(self, text):
        textList = []
        emojiList = []
        twit = Twitter()

        posText = twit.pos(text)
        posArray = np.array(posText)

        for i in range(len(posArray)):
            if posArray[i][1] == 'KoreanParticle':
                emojiList.append(posArray[i][0])

        for i in range(len(emojiList)):
            splitText = text.split(emojiList[i], maxsplit=1)

            if splitText[0] == '':
                textList.append('')
            else:
                textList.append(splitText[0])

            try:
                if len(splitText[1:]) > 1:
                    text = ''.join(splitText[1:]).strip()
                else:
                    text = splitText[1:][0].strip()

            except:
                break

            try:
                if text in emojiList[i+1]:
                    pass
            except:
                textList.append(splitText[-1])
                emojiList.append('')
                break

        ## 이모지 없는 경우            
        if len(emojiList) < 1:
            emojiList.append('')
            textList.append(text)
                
        return emojiList, textList
    




from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
import torch
from transformers import BatchEncoding

class KoTANStyleConversiontFactory:
    """
    Speech-style convert using NHNDQ/bart-speech-style-converter

    - dataset: Train

    Args:
        src (str): source language
        
    Returns:
        class: KoTANConversion class

    Examples:
        >>> mt = KoTAN(task="translation", tgt="en")
    """
    def __init__(
            self,
            task,
            tgt, 
            level,
            style
            ):
        super().__init__()
        self.style = style

    def load(self, device: str):
        tokenizer = AutoTokenizer.from_pretrained("NHNDQ/bart-speech-style-converter")
        model = AutoModelForSeq2SeqLM.from_pretrained("NHNDQ/bart-speech-style-converter").to(device)
            
        return KoTANConversion(
            model,
            tokenizer,
            device,
            self.style
        )
    

class KoTANConversion:
    def __init__(self, model, tokenizer, device, style):
        self.model = model
        self.tokenizer = tokenizer
        self.device = device
        self.style = style

    def predict(self, text):
        """
        Predict a style convert result

        Args:
            text (str): input text
            style (str): speech style

        Returns:
            output (list): Style conversion results
        """

        if type(text) == type([]):
            input_text = []
            
            for txt in text:
                input_text.append(f"{self.style} 형식으로 변환:" + txt)
        
        else:
            input_text = f"{self.style} 형식으로 변환:" + text

        
        inputs = self.tokenizer(input_text, max_length=128, return_tensors="pt")["input_ids"]
            
        convert_tokens = self.model.generate(
            inputs.to(self.device), max_length=128
        )

        output = self.tokenizer.batch_decode(convert_tokens, 
                                             skip_special_tokens=True, 
                                             clean_up_tokenization_spaces=False
                                             )
            
        return output
    

